#include "mesh.h"
#include <glm/ext.hpp>
#include <set>

#include "sparrowrenderer.h"
#include "material.h"
#include "buffer.h"

const char* const Mesh::flagStr[Mesh::NB_FLAGS] =
{
    "INDEXED",
    "TEXTURABLE",
    "INSTANCED",

    "MESH_3D",
    "MESH_2D",

    "TANGENT_SPACE",
    "DOUBLE_SIDED",
    "BILLBOARD",
    "SHADOWED",

    "COLOR_TEXTURE",
    "ALPHA_MASK",

    "PHONG",
    "DIFFUSE_TEXTURE",
    "EMISSION_TEXTURE",
    "SPECULAR_TEXTURE",
    "NORMAL_MAP",

    "BUMP_MAP"
};

Mesh::Mesh(const std::string &name) :
    m_name(name),
    material(NULL),
    isDoubleSided(false),
    isBillboard(false),
    isShadowCaster(true),
    depth(0),
    primitive_type(GL_TRIANGLES),
    vao(0)
{
    clearBuffers();
}

Mesh::~Mesh()
{
    destroyGL();
}

void Mesh::addBuffer(Buffer *b, int bufferType)
{
    buffersId[bufferType] = buffers.size();
    buffers.push_back(b);
}

void Mesh::clearBuffers()
{
    for(int i=0; i<NB_BUFFERS; ++i)
        buffersId[i] = -1;
    for(Buffer *b : buffers)
        delete(b);
    buffers.clear();
}

void Mesh::initGL()
{
    destroyGL();
    
    // create VAO
    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);

    // init positions VBO
    if(!positions3D.empty())
    {
        auto posBuffer = new TBuffer<glm::vec3>(positions3D, Buffer::VBO);
        posBuffer->setVertexAttrib(0, 3);
        addBuffer(posBuffer, POSITION3D_BUFFER);
        
        // init normals vbo
        if(!normals.empty())
        {
            auto normalsBuffer = new TBuffer<glm::vec3>(normals, Buffer::VBO);
            normalsBuffer->setVertexAttrib(1, 3);
            addBuffer(normalsBuffer, NORMAL_BUFFER);
        }
        // init tangents vbo
        if(!tangents.empty())
        {
            auto tangentsBuffer = new TBuffer<Tangents>(tangents, Buffer::VBO);
            tangentsBuffer->setVertexAttrib(3, 3, 0);
            tangentsBuffer->setVertexAttrib(4, 3, sizeof(glm::vec3));
            addBuffer(tangentsBuffer, TANGENT_BUFFER);
        }
    }
    else if(!positions2D.empty())
    {
        auto posBuffer = new TBuffer<glm::vec2>(positions2D, Buffer::VBO);
        posBuffer->setVertexAttrib(0, 2);
        addBuffer(posBuffer, POSITION2D_BUFFER);
    }
    else
    {
        fprintf(stderr, "ERROR : Mesh can't be initialized without position data");
        return;
    }
    // init texCoords vbo
    if(!texCoords.empty())
    {
        auto texCoordBuffer = new TBuffer<glm::vec2>(texCoords, Buffer::VBO);
        texCoordBuffer->setVertexAttrib(2, 2);
        addBuffer(texCoordBuffer, TEXCOORD_BUFFER);
    }
    
    // init instances vbo
    if(!instances_offsets.empty())
    {
        auto instancesBuffer = new TBuffer<glm::vec3>(instances_offsets, Buffer::VBO);
        instancesBuffer->setVertexAttrib(5, 3, 0, 1);
        addBuffer(instancesBuffer, INSTANCE_BUFFER);
    }
        
        // init EBO
    if(!indices.empty())
    {
        auto indicesBuffer = new TBuffer<GLuint>(indices, Buffer::EBO);
        addBuffer(indicesBuffer, INDICES_BUFFER);
    }

    // unbind vao
    glBindVertexArray(0);
}

void Mesh::draw(Shader* shader, bool drawNormals, bool drawTexCoord, bool drawTangents)
{
    if(isDoubleSided)
        glDisable(GL_CULL_FACE);
    if(material != NULL)
        material->bindAttributes(shader);
    glBindVertexArray(vao);
    
    if(indices.empty())
    {
        int size = positions3D.empty() ? positions2D.size() : positions3D.size();
        if(!instances_offsets.empty())
        {
            glDrawArraysInstanced(primitive_type, 0, size, instances_offsets.size());
        }
        else
        {
            glDrawArrays(primitive_type, 0, size);
        }
    }
    else
    {
        Buffer *b = buffers[buffersId[INDICES_BUFFER]];
        b->bind();
        if(!instances_offsets.empty())
        {
            glDrawElementsInstanced(primitive_type, indices.size(), GL_UNSIGNED_INT, NULL, instances_offsets.size());
        }
        else
        {
            glDrawElements(primitive_type, indices.size(), GL_UNSIGNED_INT, NULL);
        }
        b->unbind();
    }
    
    glBindVertexArray(0);
    if(isDoubleSided)
        glEnable(GL_CULL_FACE);
}

void Mesh::destroyGL()
{
    if(vao != 0)
    {
        glDeleteVertexArrays(1, &vao);
        vao = 0;
    }
    clearBuffers();
}

void Mesh::clearData()
{
    positions3D.clear();
    normals.clear();
    tangents.clear();
    positions2D.clear();
    texCoords.clear();
    instances_offsets.clear();
    indices.clear();
}

unsigned int Mesh::getFlags()
{
    unsigned int flags = material->getFlags();

    if(!(indices.empty() || buffersId[INDICES_BUFFER] == -1))
        flags |= 1 << MESH_INDEXED;
    if(!(texCoords.empty() || buffersId[TEXCOORD_BUFFER] == -1))
        flags |= 1 << MESH_TEXTURABLE;
    if(!(instances_offsets.empty() || buffersId[INSTANCE_BUFFER] == -1))
        flags |= 1 << MESH_INSTANCED;

    if(!(positions3D.empty() || buffersId[POSITION3D_BUFFER] == -1))
    {
        flags |= 1 << MESH_3D;

        if((!tangents.empty() || buffersId[TANGENT_BUFFER] == -1))
            flags |= 1 << MESH_TANGENT_SPACE;
        if(isDoubleSided)
            flags |= 1 << MESH_DOUBLE_SIDED;
        if(isBillboard)
            flags |= 1 << MESH_BILLBOARD;
        if(isShadowCaster)
            flags |= 1 << MESH_SHADOWED;
    }
    else
        flags |= 1 << MESH_2D;

    return flags;
}

void Mesh::addRectangle2D(const glm::vec2 &pos, const glm::vec2 &dim, const glm::vec2 &texCoord, const glm::vec2 &texRange, bool indexed)
{
    if(indexed)
    {
        addVertex(pos,                     texCoord);
        addVertex(pos+glm::vec2(dim.x, 0), texCoord+glm::vec2(texRange.x, 0));
        addVertex(pos+dim,                 texCoord+texRange);
        addVertex(pos+glm::vec2(0, dim.y), texCoord+glm::vec2(0, texRange.y));
        
        addTriangle(0, 1, 2);
        addTriangle(2, 3, 0);
    }
    else
    {
        addVertex(pos,                     texCoord);
        addVertex(pos+glm::vec2(dim.x, 0), texCoord+glm::vec2(texRange.x, 0));
        addVertex(pos+dim,                 texCoord+texRange);
        
        addVertex(pos+dim,                 texCoord+texRange);
        addVertex(pos+glm::vec2(0, dim.y), texCoord+glm::vec2(0, texRange.y));
        addVertex(pos,                     texCoord);
    }
}

struct VertexComparator
{
    // c'est plutot crade mais j'ai pas trouve d'autre moyen pour le moment
    static Mesh* mesh;
    static void setMesh(Mesh* m) {VertexComparator::mesh = m;}

    bool operator() (const int& vertId1, const int& vertId2) const
    {
        if(mesh->positions3D[vertId1].x != mesh->positions3D[vertId2].x)
            return (mesh->positions3D[vertId1].x < mesh->positions3D[vertId2].x);
        if(mesh->positions3D[vertId1].y != mesh->positions3D[vertId2].y)
            return (mesh->positions3D[vertId1].y < mesh->positions3D[vertId2].y);
        if(mesh->positions3D[vertId1].z != mesh->positions3D[vertId2].z)
            return (mesh->positions3D[vertId1].z < mesh->positions3D[vertId2].z);
        if(!mesh->texCoords.empty())
        {
            if(mesh->texCoords[vertId1].x - floor(mesh->texCoords[vertId1].x) != mesh->texCoords[vertId2].x - floor(mesh->texCoords[vertId2].x))
                return (mesh->texCoords[vertId1].x < mesh->texCoords[vertId2].x);
            if(mesh->texCoords[vertId1].y - floor(mesh->texCoords[vertId1].y) != mesh->texCoords[vertId2].y - floor(mesh->texCoords[vertId2].y))
                return (mesh->texCoords[vertId1].y < mesh->texCoords[vertId2].y);
        }
        if(!mesh->normals.empty())
        {
            if(mesh->normals[vertId1].x != mesh->normals[vertId2].x)
                return (mesh->normals[vertId1].x < mesh->normals[vertId2].x);
            if(mesh->normals[vertId1].y != mesh->normals[vertId2].y)
                return (mesh->normals[vertId1].y < mesh->normals[vertId2].y);
            if(mesh->normals[vertId1].z != mesh->normals[vertId2].z)
                return (mesh->normals[vertId1].z < mesh->normals[vertId2].z);
        }
        return false;
    }
};

Mesh* VertexComparator::mesh = NULL;

void Mesh::mergeVertices()
{
    if(positions3D.empty())
        return;
    bool *deleted = new bool[positions3D.size()];
    int *offsets = new int[positions3D.size()];
    std::set<int, VertexComparator> vertexSet;
    VertexComparator::setMesh(this);

    for(std::size_t i=0; i<indices.size(); ++i)
    {
        std::pair<std::set<int,VertexComparator>::iterator, bool> ret = vertexSet.insert(indices[i]);
        std::set<int,VertexComparator>::iterator it = ret.first;
        bool success = ret.second;
        deleted[indices[i]] = !success && *it != int(indices[i]);
        if(deleted[indices[i]])
        {
            if(!tangents.empty())
                tangents[*it].tangent += tangents[indices[i]].tangent;
            indices[i] = *it;
        }
    }
    int offset = 0;
    int pos = 0;
    for(std::size_t i=0; i<positions3D.size(); ++i)
    {
        if(deleted[i])
            ++offset;
        else
        {
            offsets[i] = offset;
            if(offset != 0)
            {
                positions3D[pos] = positions3D[i];
                if(!texCoords.empty())
                    texCoords[pos] = texCoords[i];
                if(!normals.empty())
                    normals[pos] = normals[i];
                if(!tangents.empty())
                    tangents[pos] = tangents[i];
            }
            ++pos;
        }
    }
    for(std::size_t i=0; i<indices.size(); ++i)
        indices[i] -= offsets[indices[i]];

    positions3D.resize(positions3D.size()-offset);
    if(!texCoords.empty())
        texCoords.resize(texCoords.size()-offset);
    if(!normals.empty())
        normals.resize(normals.size()-offset);
    if(!tangents.empty())
        tangents.resize(tangents.size()-offset);
    for(int i=0; i<int(tangents.size()); ++i)
    {
        glm::vec3 &T = tangents[i].tangent;
        const glm::vec3 &N = normals[i];
        
        T = glm::normalize(T);
        // re-orthogonalize T with respect to N
        T = glm::normalize(T - glm::dot(T, N) * N);
        // then retrieve perpendicular vector B with the cross product of T and N
        tangents[i].binormal = glm::normalize(glm::cross(T, N));
    }
}

void Mesh::computeNeighbors()
{
    if(positions3D.empty())
        return;
    // TODO : compute adjacency and change primitivetype
}

void Mesh::computeNormals()
{
    if(positions3D.empty())
        return;
    normals.resize(positions3D.size());
    std::memset(normals.data(), 0, normals.size());
    for (std::size_t i=0; i < indices.size(); i += 3)
    {
        int v0 = indices[i];
        int v1 = indices[i+1];
        int v2 = indices[i+2];
        glm::vec3 n = glm::cross(positions3D[v1] - positions3D[v0], positions3D[v2] - positions3D[v0]);
        normals[v0] += n;
        normals[v1] += n;
        normals[v2] += n;
    }
    for(glm::vec3 &n : normals)
        n = glm::normalize(n);
}

void Mesh::computeTangents()
{
    if(texCoords.empty())
        return;
    tangents = std::vector<Tangents>(positions3D.size());

    for (std::size_t j=0; j < indices.size(); j += 3)
    {
        int vertexId0 = indices[j];
        int vertexId1 = indices[j+1];
        int vertexId2 = indices[j+2];

        const glm::vec3 &v1 = positions3D[vertexId0];
        glm::vec3 edge1 = positions3D[vertexId1] - v1;
        glm::vec3 edge2 = positions3D[vertexId2] - v1;
        
        const glm::vec2& w1 = texCoords[vertexId0];
        glm::vec2 deltaUV1 = texCoords[vertexId1] - w1;
        glm::vec2 deltaUV2 = texCoords[vertexId2] - w1;
        
        float f = 1.0f / (deltaUV1.x * deltaUV2.y - deltaUV2.x * deltaUV1.y);
        
        glm::vec3 tangentDir = deltaUV2.y * edge1 - deltaUV1.y * edge2;
        tangentDir = glm::normalize(tangentDir*f);
        
        glm::vec3 binormalDir = deltaUV1.x * edge2 - deltaUV2.x * edge1;
        binormalDir = glm::normalize(binormalDir*f);
        if(glm::dot(glm::cross(normals[vertexId0], binormalDir), tangentDir) < 0.f)
            binormalDir = -binormalDir;
        
        tangents[vertexId0] = {tangentDir, binormalDir};
        tangents[vertexId1] = {tangentDir, binormalDir};
        tangents[vertexId2] = {tangentDir, binormalDir};
    }
}

void Mesh::computeBoundingBox(glm::vec3 &min, glm::vec3 &max)
{
    min = max = positions3D[0];
    for(const glm::vec3 &pos : positions3D)
    {
        if(pos.x < min.x)
            min.x = pos.x;
        else if(pos.x > max.x)
            max.x = pos.x;
        if(pos.y < min.y)
            min.y = pos.y;
        else if(pos.y > max.y)
            max.y = pos.y;
        if(pos.z < min.z)
            min.z = pos.z;
        else if(pos.z > max.z)
            max.z = pos.z;
    }
}

// serialisation methods

struct MeshHeader
{
    unsigned int flags;
    int nbPositions;
    int depth;
    int nbInstances_offsets;
    int nbIndices;
    int nameLength;
};

template<typename T>
bool writeBuffer(const std::vector<T> &vec, std::FILE *file)
{
    size_t nbWritten = std::fwrite(vec.data(), sizeof(T), vec.size(), file);
    return (nbWritten == vec.size());
}

template<typename T>
bool readBuffer(std::vector<T> &vec, std::FILE *file)
{
    size_t nbRead = std::fread(vec.data(), sizeof(T), vec.size(), file);
    return (nbRead == vec.size());
}

bool Mesh::serialize(Mesh* mesh, FILE *file)
{
    // creating header
    MeshHeader header;
    header.flags = mesh->getFlags();
    if(header.flags & (1 << Mesh::MESH_3D))
        header.nbPositions = mesh->positions3D.size();
    else
    {
        header.nbPositions = mesh->positions2D.size();
        header.depth = mesh->getDepth();
    }
    header.nbIndices = mesh->indices.size();
    header.nbInstances_offsets = mesh->instances_offsets.size();
    header.nameLength = mesh->getName().size();
    
    if(header.nbPositions == 0)
        return false;
    
    // writing header
    size_t nbWritten;
    
    nbWritten = std::fwrite(&header, sizeof(MeshHeader), 1, file);
    if(nbWritten != 1)
        return false;
    
    if(header.nameLength != 0)
    {
        nbWritten = std::fwrite(mesh->getName().data(), header.nameLength, 1, file);
        if(nbWritten != 1)
            return false;
    }

    // writing buffers
    if(header.flags & (1 << Mesh::MESH_3D))
    {
        if(!writeBuffer(mesh->positions3D, file))
            return false;
        if(mesh->normals.size() == 0)
            mesh->computeNormals();
        if(!writeBuffer(mesh->normals, file))
            return false;
        if(header.flags & (1 << Mesh::MESH_TANGENT_SPACE))
        {
            if(!writeBuffer(mesh->tangents, file))
                return false;
        }
    }
    else
    {
        if(!writeBuffer(mesh->positions2D, file))
            return false;
    }
    
    if(header.nbInstances_offsets)
    {
        if(!writeBuffer(mesh->instances_offsets, file))
            return false;
    }
    
    if(header.nbIndices)
    {
        if(!writeBuffer(mesh->indices, file))
            return false;
    }
    
    if(header.flags & (1 << Mesh::MESH_TEXTURABLE))
    {
        if(!writeBuffer(mesh->texCoords, file))
            return false;
    }
    
    return true;
}

Mesh* Mesh::deserialize(FILE *file)
{
    MeshHeader header;
    size_t nbRead = std::fread(&header, sizeof(MeshHeader), 1, file);
    
    // deserializing mesh
    Mesh *mesh = NULL;
    if(nbRead == 1 && header.nbPositions != 0)
        mesh = new Mesh();
    else
        return NULL;
    
    std::string name = "mesh";
    if(header.nameLength != 0)
    {
        name.resize(header.nameLength);
        if(!fread(&(name[0]), header.nameLength, 1, file))
            { delete(mesh); return NULL; }
    }
    mesh->setName(name);
    
    if(header.flags & (1 << Mesh::MESH_3D))
    {
        mesh->positions3D.reserve(header.nbPositions);
        if(!readBuffer(mesh->positions3D, file))
            { delete(mesh); return NULL; }
        mesh->normals.reserve(header.nbPositions);
        if(!readBuffer(mesh->normals, file))
            { delete(mesh); return NULL; }
        if(header.flags & (1 << Mesh::MESH_TANGENT_SPACE))
        {
            mesh->tangents.reserve(header.nbPositions);
            if(!readBuffer(mesh->tangents, file))
                { delete(mesh); return NULL; }
        }
    }
    else
    {
        mesh->positions2D.reserve(header.nbPositions);
        if(!readBuffer(mesh->positions2D, file))
            { delete(mesh); return NULL; }
        mesh->setDepth(header.depth);
    }
    
    if(header.nbInstances_offsets)
    {
        mesh->instances_offsets.reserve(header.nbInstances_offsets);
        if(!readBuffer(mesh->instances_offsets, file))
            { delete(mesh); return NULL; }
    }
    
    if(header.nbIndices)
    {
        mesh->indices.reserve(header.nbIndices);
        if(!readBuffer(mesh->indices, file))
            { delete(mesh); return NULL; }
    }
    
    if(header.flags & (1 << Mesh::MESH_TEXTURABLE))
    {
        mesh->texCoords.reserve(header.nbPositions);
        if(!readBuffer(mesh->texCoords, file))
            { delete(mesh); return NULL; }
    }
    
    mesh->setIsBillboard(header.flags & (1 << Mesh::MESH_BILLBOARD));
    mesh->setIsDoubleSided(header.flags & (1 << Mesh::MESH_DOUBLE_SIDED));
    mesh->setIsShadowCaster(header.flags & (1 << Mesh::MESH_SHADOWED));
    
    return mesh;
}
